from zipline.pipeline.factors import ExponentialWeightedMovingAverage, RSI

from bt.pipeline_factors.custom_factors import *
from bt.strategies.abs_strategy import Strategy
from zipline.api import attach_pipeline, pipeline_output, record, symbol
from zipline.pipeline import CustomFactor, Pipeline
from zipline.pipeline.data import USEquityPricing
from bt.backtest.cnn.data_retrieve import CSVWriter
from bt.backtest.cnn.normalizer import normalize_file
from zipline.utils.calendars import get_calendar
import skimage.measure

import numpy as np
import pandas as pd
import copy as cp
import math


def generate(assets, strategies, indicators, num_intervals, type, dir, categories=None, pooling=False, append=False):
    # currently expects strategies to have the correct starting and ending dates for the first interval
    start_params = cp.deepcopy(strategies[0].params)
    price_days = start_params.td
    print(price_days)
    generator_params = cp.deepcopy(start_params)

    csv_writer = CSVWriter(dir=dir)
    if(append == False):
        open(dir, 'w').close()
        csv_writer.initCSV(strategies, indicators, price_days, type, categories, pooling)

    for asset in assets:
        for strategy in strategies:
            strategy.params.update(start=cp.deepcopy(start_params.start), td=price_days, asset=asset)
        generator = GeneratorStrategy(cp.deepcopy(generator_params), asset, indicators)
        for i in range(0, num_intervals):
            print("Asset: {}, Generator start: {}, Strategy start: {}".format(strategies[0].params.asset, generator.params.start, strategies[0].params.start))
            generator.run_backtest()
            for strategy in strategies:
                strategy.run_backtest()
                #print strategy.result
                strategy.params.increment()
            performance = evaluate_performance([strategy.result for strategy in strategies])
            for strategy in strategies:
                strategy.run_backtest()
            weights = evaluate_weights([strategy.result for strategy in strategies], type, categories)
            data = np.empty((0,0))
            #Date
            data = np.append(data, [str(generator.params.start)])
            #Price Data
            price_data = convert_to_angles(generator.result['Price'].as_matrix())
            if(pooling):
                price_data = pool_data(price_data)
            data = np.append(data, price_data)
            #Indicators(Zipline Pipeline)
            for indicator_str in indicators:
                data = np.append(data, generator.indicator_values[indicator_str])
            #Performance(Inputs)
            data = np.append(data, performance)
            #Weights(Outputs)
            data = np.append(data, [weights])
            #note: it might make more sense to append asset to the data
            csv_writer.appendToCSV(data, asset)
            generator.params.increment()

    # lazy normalization
    print("Beginning normalization")
    num_lines = sum(1 for l in open(dir))
    skip_idx = [x for x in range(1, num_lines) if x > 1]
    header_df = pd.read_csv(dir, delimiter=',', delim_whitespace=False, header=0, skiprows=skip_idx)
    df = normalize_file(dir, indicators)
    header_df.to_csv(dir, index=False)
    df.to_csv(dir, mode='a', index=False)
    print("Normalization complete")


###############################################
######GENERATOR/PIPELINE CODE
##############################################


class AssetMaskFactor(CustomFactor):

    inputs = [USEquityPricing.close]
    window_length = 14
    asset = None

    def compute(self, today, assets, out, close):
        out[:] = [1 if asset == self.asset else 0 for asset in assets]

class GeneratorStrategy(Strategy):

    def __init__(self, params, asset, indicators):
        super(GeneratorStrategy, self).__init__('generator', params)
        self.asset = asset
        self.indicators = indicators
        self.indicator_values = None
        self.prices = []
        self.pipeline = None

    def initialize(self):
        def initialize(context):
            self.indicator_values = {}
            context.asset = symbol(self.asset)
            context.pipeline = self.pipeline = Pipeline()
            attach_pipeline(context.pipeline, 'pipeline')
            maskFactor = AssetMaskFactor()
            maskFactor.asset = symbol(self.asset)
            mask = (maskFactor > 0)
            context.pipeline.set_screen(mask)
            for indicator_str in self.indicators:
                factor = None
                if indicator_str == 'EMA':
                    factor = ExponentialWeightedMovingAverage(inputs=[USEquityPricing.close], window_length=14, decay_rate=1.0)
                elif indicator_str == 'LongEMA':
                    factor = ExponentialWeightedMovingAverage(inputs=[USEquityPricing.close], window_length=42, decay_rate=1.0)
                elif indicator_str == 'RSI':
                    factor = RSI()
                elif indicator_str == 'PercentK':
                    factor = PercentK(14)
                elif indicator_str == 'PercentD':
                    factor = PercentD(14, 3)
                elif indicator_str == 'SlowPercentD':
                    factor = SlowPercentD(14, 3, 3)
                elif indicator_str == 'Momentum':
                    factor = Momentum(4)
                elif indicator_str == 'ROC':
                    factor = ROC(4)
                elif indicator_str == 'PercentR':
                    factor = PercentR(14)
                elif indicator_str == 'AccumulationDistribution':
                    factor = AccumulationDistribution(252) #needs around a year of recursion for accuracy
                elif indicator_str == 'ChalkinOscillator':
                    factor = ChalkinOscillator(252) #uses accumulationdistribution
                elif indicator_str == 'CommoditiesChannelIndex':
                    factor = CommoditiesChannelIndex(7)
                elif indicator_str == 'Disparity5':
                    factor = Disparity(5)
                elif indicator_str == 'Disparity10':
                    factor = Disparity(10)
                elif indicator_str == 'Votality':
                    factor = Votality(14)
                else:
                    continue
                factor.mask = mask
                context.pipeline.add(factor, indicator_str)

        return initialize

    def handle_data(self):
        def handle_data(context, data):
            record(Price=data.current(context.asset, 'price'))
        return handle_data

    def analyze(self):
        def analyze(context, perf):
            for indicator_str in self.indicators:
                self.indicator_values[indicator_str] = pipeline_output('pipeline')[indicator_str].data[0]
            #print pipeline_output('pipeline')
            #print self.indicator_values
        return analyze

#########################################################
#####EVALUATION FUNCTIONS
#########################################################

def evaluate_performance(results):
    return complex_performance_eval(results)


def evaluate_weights(results, type, categories):
    if type == 'regression':
        return reg_weight_eval(results)
    elif type == 'classification':
        return cat_weight_eval(results, categories)


def cat_weight_eval(results, num_cat):
    performance = np.empty((0, 0))
    bins = np.linspace(-1, 1, num_cat + 1)
    for result in results:
        print(result)
        principal, final = result['starting_cash'][0], result['starting_cash'][-1]
        print(final, principal)
        percent_change = (float(final - principal) / float(principal)) * 100.0
        #NOTE: CHANGE THE FOLLOWING VALUE BASED ON THE TOLERANCE YOU WANT FOR THE NETWORK
        range = 3.0
        score = round(force_within_range(percent_change / range, -0.99, 0.99),2)
        #experiment with right
        score = np.digitize(score, bins, right=True) - 1
        print("Score output: {}".format(score))
        performance = np.append(performance, [score])
    return performance


def reg_weight_eval(results):
    performance = np.empty((0, 0))
    for result in results:
        principal, final = result['starting_cash'][0], result['starting_cash'][-1]
        percent_change = (float(final - principal) / float(principal)) * 100.0
        score = round(force_within_range(percent_change / 10.0, -1, 1),2)
        performance = np.append(performance, [score])
    return performance


def complex_performance_eval(results):
    tc = get_calendar('NYSE')
    performance = np.empty((0,0))
    for result in results:
        result = reindex(result)
        principal = result['starting_cash'][0]
        final = result['starting_cash'][-1]
        trade_dates = [(result['trades'][i], result['date'][i]) \
                       for i in range(0, len(result.index)) if check_clean(result['trades'][i])]
        winning_trades, losing_trades = np.empty((0,4)), np.empty((0,4))
        num_invalid_kratio = 0

        for trade in trade_dates:
            trade = (trade[0], tc.next_open(tc.next_open(trade[1])).replace(hour=0, minute=0))
            diff = (result['starting_cash'][trade[1]] - result['starting_cash'][trade[0]]) / result['starting_cash'][trade[0]]
            portfolio_value = result['starting_value'][trade[0]:trade[1]] + result['starting_cash'][trade[0]:trade[1]]
            drawback = calculate_drawback(portfolio_value)
            k_ratio = calculate_kratio(portfolio_value)
            if k_ratio is None:
                num_invalid_kratio += 1
                k_ratio = 0
            if diff >= 0:
                winning_trades = np.append(winning_trades, [[trade, diff, drawback, k_ratio]], axis=0)
            else:
                losing_trades = np.append(losing_trades, [[trade, diff, drawback, k_ratio]], axis=0)

        win_count, loss_count = winning_trades.shape[0], losing_trades.shape[0]
        percent_return = (float(final - principal) / float(principal)) * 10
        #percent_return = force_within_range(percent_return, -1, 1)

        if win_count == 0 and loss_count == 0:
            win_ratio,win_percent,loss_percent,average_drawdown,average_kratio = 0,0,0,0,0
        else:
            if win_count == 0:
                win_percent = 0
                loss_percent = np.sum(losing_trades[:, 1]) / losing_trades.shape[0]
            elif loss_count == 0:
                win_percent = np.sum(winning_trades[:, 1]) / winning_trades.shape[0]
                loss_percent = 0
            else:
                win_percent = np.sum(winning_trades[:, 1]) / winning_trades.shape[0]
                loss_percent = np.sum(losing_trades[:, 1]) / losing_trades.shape[0]

            win_ratio = (float)(winning_trades.shape[0]) / (float)(winning_trades.shape[0] + losing_trades.shape[0])

            average_drawdown = (np.sum(losing_trades[:, 2]) + np.sum(winning_trades[:, 2])) \
                                / (winning_trades.shape[0] + losing_trades.shape[0])

            average_kratio = (np.sum(winning_trades[:, 3]) + np.sum(losing_trades[:, 3])) \
                                / (winning_trades.shape[0] + losing_trades.shape[0] - num_invalid_kratio)

            loss_percent = abs(loss_percent)
        #print("Percent Return: ", percent_return)
        #print("Win Ratio: ", win_ratio)
        #print("Win Percent: ", win_percent)
        #print("Loss Percent: ", loss_percent)
        #print("Average Drawdown: ", average_drawdown)
        #print("Average KRatio: ", average_kratio)

        performance = np.append(performance, np.array([percent_return, win_ratio, win_percent, loss_percent, average_drawdown, average_kratio]))
    return performance


###########################################################
####HELPERS
###########################################################

def pool_data(price_data):
    reduced = skimage.measure.block_reduce(price_data, (4,), np.mean)
    return reduced

def convert_to_angles(raw_data):
    diff = raw_data[1:] - raw_data[:-1]
    angles = np.arctan(diff) / (np.pi / 2)
    return np.round(angles, 3)

def check_clean(time):
    return time is not None and not pd.isnull(time)

def reindex(result):
    result.index = result.index.map(lambda t: t.replace(hour=0, minute=0))
    return result

def calculate_drawback(prices):
    best = 0
    best_mx = prices[0]
    mx = prices[0]
    mn = prices[0]
    for elt in prices:
        if elt > mx:
            if mx - mn > best:
                best = mx - mn
                best_mx = mx
            mx = elt
            mn = elt
        if elt < mn:
            mn = elt
    if mx-mn > best:
        best = mx-mn
        best_mx = mx
    return best / best_mx


def calculate_kratio(prices):
    y = prices.as_matrix()
    y = np.reshape(y, (1, y.shape[0]))
    x = np.arange(0, y.shape[1])
    x = np.reshape(x, (1, y.shape[1]))
    if x.shape[1] > 2:
        lin_reg = np.polyfit(x[0], y[0], 1)
        vals = lin_reg[0] * x + lin_reg[1] * np.ones((1, x.shape[1]))
        error = math.sqrt(np.sum(np.square(y - vals))) / (x.shape[1] - 2)
        return lin_reg[0]/error
        #plt.plot(np.unique(x), np.poly1d(lin_reg)(np.unique(x)))
    else:
        return None # kratio cannot be calculated if there are less than 3 trades

def force_within_range(num, min, max):
    if num <= min:
        return min
    elif num >= max:
        return max
    else:
        return num