import zipline
from zipline.utils.calendars import get_calendar
from datetime import timedelta
from zipline.api import (
    order_percent,
    order_value,
    symbol,
    order_target_percent,
    order_target_value,
    order_target,
    record,
    get_order,
    get_environment,
    get_datetime,
    cancel_order
)
import numpy as np
from bt.strategies.strategy_params import StrategyParameters

class HelperContext():

    def __init__(self, asset):
        self.i = 0
        self.asset = symbol(asset)
        self.buy_id = None
        self.sell_id = None
        self.buy_date= None
        self.at_sell = None
        self.stop_orders = False


class Strategy(object):

    def __init__(self, name, params):
        self.name = name
        self.params = params
        self.result = None
        pass

    def initialize(self):
        def initialize(context):
            pass
        return initialize

    def handle_data(self):
        def handle_data(context, data):
            pass
        return handle_data

    def analyze(self):
        def analyze(context, perf):
            pass
        return analyze

    def set_start(self, year, month, day):
        self.params.update(start=StrategyParameters.parse_time(year, month, day))

    def set_end(self, year, month, day):
        self.params.update(end=StrategyParameters.parse_time(year, month, day))

    def set_principal(self, principal):
        self.params.update(principal=principal)

    def set_frequency(self, frequency):
        self.params.update(frequency=frequency)

    def set_bundle(self, bundle):
        self.params.update(bundle=bundle)

    def set_asset(self, asset):
        self.params.update(asset=asset)

    def increment_timedelta(self, principal=None):
        self.params.increment(principal=principal)

    def run_backtest(self, params=None):
        if not params:
            params = self.params
        result = zipline.run_algorithm(params.start, params.end, self.initialize(), params.principal, self.handle_data(),
                              analyze=self.analyze(), data_frequency=params.frequency, bundle=params.bundle)

        self.result = result
        return result

    def run_backtest(self, start = None, end = None, principal = None, frequency = None, bundle = None, asset = None):

        if not start:
            start = self.params.start
        if not end:
            end = self.params.end
        if not principal:
            principal = self.params.principal
        if not frequency:
            frequency = self.params.frequency
        if not bundle:
            bundle = self.params.bundle
        if not asset:
            asset = self.params.asset

        result = zipline.run_algorithm(start, end, self.initialize(), principal, self.handle_data(),
                                       analyze=self.analyze(), data_frequency=frequency, bundle=bundle)
        #TODO: figure out how to reenamble auto-updates
        #self.params.update(start=start, ts=end-start, principal=principal, frequency=frequency, bundle=bundle)

        self.result = result
        return result

    def wait_bars_back(self, context, bb):
        context.object.i += 1
        if context.object.i < bb:
            return False
        return True

    def nn_record(self, context, data):
        tc = get_calendar('NYSE')
        if context.object.buy_id is not None:
            #TODO: switch to status maybe
            if get_order(context.object.buy_id)['filled'] != 0:
                context.object.buy_date = get_datetime().replace(hour=0,minute=0)
                context.object.buy_id = None
        if context.object.sell_id is not None:
            if get_order(context.object.sell_id)['filled'] != 0:
                context.object.at_sell = context.object.buy_date
                context.object.buy_date = None
                context.object.sell_id = None
        record(date=get_datetime().replace(hour=0, minute=0),
               price=data.current(context.object.asset, "price"), #for debug purposes only
               trades=context.object.at_sell)
        if context.object.at_sell is not None:
            context.object.at_sell = None

    def order_principal(self, context, data):
        principal = get_environment('capital_base')
        if self.get_cash_balance(context) >= principal:
            order_target_value(context.object.asset, principal)
        else:
            order_target_percent(context.object.asset, 1)

    def sell(self, context, data):
        context.object.sell_id = order_target_percent(context.object.asset, 0)

    def test_buy(self, context, data):
        context.object.buy_id = order_target(context.object.asset, 8)

    def test_sell(self, context, data):
        context.object.sell_id = order_target(context.object.asset, 0)

    def cancel_orders(self, context, data):
        [[cancel_order(order) for order in val] for key, val in context.get_open_orders().items()]

    def get_cash_balance(self, context):
        return context.portfolio['cash']

    def get_positions_exposure(self, context):
        return context.portfolio['positions_exposure']

    def get_portfolio_value(self, context):
        return context.portfolio['portfolio_value']

    def get_portfolio_return(self, context):
        return context.portfolio['returns']

    def sell_at_end(self, context, data):
        pass

    def sell_at_runoff(self, context, data):
        #TODO implement this
        pass

    def set_trading_period(self, context, elts):
        pass

    def check_trading_period(self, context):
        pass


class DailyStrategy(Strategy):

    def __init__(self, name, params):
        super(DailyStrategy, self).__init__(name, params)

    def sell_at_end(self, context, data):
        tc = get_calendar('NYSE')
        a = tc.next_open(tc.next_open(get_datetime())).replace(hour=0,minute=0)
        b = tc.previous_open(tc.previous_open(get_environment('end'))).replace(hour=0,minute=0)
        #a = tc.next_open(get_datetime()).replace(hour=0, minute=0)
        #b = tc.previous_open(get_environment('end')).replace(hour=0,minute=0)
        if a == b:
            self.cancel_orders(context, data)
            self.sell(context, data)
            context.object.stop_orders = True

    def set_trading_period(self, context, elts):
        context.object.period = elts

    #THIS MIGHT BE BROKEN
    def check_trading_period(self, context):
        return (get_datetime().day % context.object.period) == 0


class MinuteStrategy(Strategy):

    def __init__(self, name, params):
        super(MinuteStrategy, self).__init__(name, params)

    def sell_at_end(self, context, data):
        tc = get_calendar('NYSE')
        if tc.next_minute(tc.next_minute(get_datetime())) == get_environment('end') - timedelta(minutes=3):
            self.cancel_orders(context, data)
            self.sell(context, data)

    def set_trading_period(self, context, elts):
        context.object.period = elts

    def check_trading_period(self, context):
        return (get_datetime().minute % context.object.period) == 0