from bt.strategies.abs_strategy import *
from bt.strategies.strategy_params import *
from math import isnan
from pandas import DataFrame
from zipline.api import (
    order,
    record,
    symbol,
    set_slippage,
    set_commission
)
from zipline.utils.calendars import get_calendar
import zipline
# from helpers import *
from datetime import timedelta
import unittest


class VIXTimingStrategy(DailyStrategy):

    def __init__(self, start, end, principal, frequency, bundle, asset, name='VIXTimingStrategy'):
        params = DailyStrategyParameters(start, end,
                                         principal, frequency, bundle, asset)
        super(VIXTimingStrategy, self).__init__(name, params)

    def initialize(self):
        def initialize(context):
            context.object = HelperContext(self.params.asset)
            context.order_price = 0
            context.has_ordered = False
            set_commission(us_equities=None, us_futures=None)
            set_slippage(us_equities=None, us_futures=None)
            pass

    def handle_data(self):
        def handle_data(context, data):

            self.wait_bars_back(context, 10)

            last10low = data.history(context.object.asset, "low", 10, "1d")
            last10close = data.history(context.object.asset, "close", 10, "1d")
            last10high = data.history(context.object.asset, "high", 10, "1d")

            open = data.current(context.object.asset, "open")
            close = data.current(context.object.asset, "close")
            low = data.current(context.object.asset, "low")
            high = data.current(context.object.asset, "high")

            if not context.has_ordered:
                check1 = low <= last10low.mean()
                check2 = close < open
                check3 = (close - last10close.mean()) >= (last10close.mean() / 10)

                if check1 and check2 and check3:
                    self.test_buy(context, data)
                    context.has_ordered = True

            elif context.has_ordered:
                check1 = high <= last10high.mean()
                check2 = (last10close.mean() - close) >= (last10close.mean() / 10)
                check3 = close > open

                if check1 and check2 and check3:
                    self.sell(context, data)
                    context.has_ordered = False

            self.sell_at_end(context, data)
            self.nn_record(context, data)

        return handle_data


class SimpleEMAStrategy(DailyStrategy):

    def __init__(self, start, end, principal, frequency, bundle, asset, name='SimpleEMAStrategy'):
        params = DailyStrategyParameters(start, end,
                                         principal, frequency, bundle, asset)
        super(SimpleEMAStrategy, self).__init__(name, params)

    def initialize(self):
        def initialize(context):
            context.object = HelperContext(self.params.asset)
            context.order_price = 0
            context.has_ordered = False
            set_commission(us_equities=None, us_futures=None)
            set_slippage(us_equities=None, us_futures=None)
            pass

        return initialize

    def handle_data(self):
        def handle_data(context, data):

            self.wait_bars_back(context, 10)

            if not context.object.stop_orders:

                last10 = data.history(context.object.asset, "high", 10, "1d")
                price = data.current(context.object.asset, "price")

                if last10.max() == last10[-1] and not context.has_ordered:
                    # buy_max_starting(context, data)
                    # order_percent(context.asset, 1)
                    self.test_buy(context, data)
                    context.has_ordered = True
                    context.order_price = price

                elif context.has_ordered:
                    returns = data.current(context.object.asset, "price") - context.order_price
                    if (abs(returns) > 3):
                        # sell_max_starting(context, data)
                        # order_percent(context.asset, -1)
                        self.sell(context, data)
                        context.has_ordered = False

                self.sell_at_end(context, data)
            self.nn_record(context, data)

        return handle_data


class TurtleshStrategy(DailyStrategy):
    CONST_LONG = 1
    CONST_SHORT = -1
    DEBUG = False

    def __init__(self, start, td, principal, frequency, bundle, asset, name='TurtleshStrategy'):
        params = DailyStrategyParameters(start, td, principal, frequency, bundle, asset)
        super(TurtleshStrategy, self).__init__(name, params)

    def initialize(self):
        def initialize(context):
            context.first_run = True
            context.object = HelperContext(self.params.asset)
            context.pdn = None
            context.strat = None
            context.stop_price = None
            context.units_in = 0

        return initialize

    def handle_data(self):
        def handle_data(context, data):
            if not context.object.stop_orders:
                N = TurtleshStrategy.n(context, data)

                unit = 0.1 * zipline.api.get_environment(field='capital_base') / N
                if isnan(unit): # I don't know why this would happen, but it seems to be happening...
                    print("unit is NaN!")
                    unit = 0
                context.pdn = N

                price_low = data.current(context.object.asset, 'low')
                price_high = data.current(context.object.asset, 'high')

                # Entries
                if price_high >= data.history(context.object.asset, 'high', 55, "1d").max() and context.strat is None:
                    # long 1 unit
                    if TurtleshStrategy.DEBUG:
                        print("Longing 1 unit")
                    context.object.buy_id = zipline.api.order_value(context.object.asset, unit)
                    context.stop_price = price_high - 2 * N
                    context.strat = TurtleshStrategy.CONST_LONG
                    context.units_in = 1
                elif context.strat == TurtleshStrategy.CONST_LONG and context.units_in < 4 and price_high >= data.history(
                        context.object.asset,
                        'high', 55,
                        "1d").max():
                    # long additional unit
                    if TurtleshStrategy.DEBUG:
                        print("Longing an additional unit")
                    context.units_in += 1
                    context.object.buy_id = zipline.api.order_value(context.object.asset, context.units_in * unit)
                    context.stop_price += 0.5 * N
                elif price_low <= data.history(context.object.asset, 'low', 55, "1d").min() and context.strat is None:
                    # short 1 unit
                    if TurtleshStrategy.DEBUG:
                        print("Shorting 1 unit")
                    context.object.buy_id = zipline.api.order_value(context.object.asset, -1 * unit)
                    context.stop_price = price_low + 2 * N
                    context.strat = TurtleshStrategy.CONST_SHORT
                    context.units_in = 1
                elif context.strat == TurtleshStrategy.CONST_SHORT and context.units_in < 4 and price_low <= data.history(
                        context.object.asset,
                        'low', 55,
                        "1d").min():
                    # short additional unit
                    if TurtleshStrategy.DEBUG:
                        print("Shorting an additional unit")
                    context.units_in += 1
                    context.object.buy_id = zipline.api.order_value(context.object.asset, -1 * context.units_in * unit)
                    context.stop_price -= 0.5 * N

                # Stops
                if (context.strat == TurtleshStrategy.CONST_LONG and price_low < context.stop_price) or (
                        context.strat == TurtleshStrategy.CONST_SHORT and price_high > context.stop_price):
                    if TurtleshStrategy.DEBUG:
                        print("Stopping...")
                    context.object.sell_id = zipline.api.order_target(context.object.asset, 0)
                    context.strat = None

                # Exits
                if (price_low <= data.history(context.object.asset, 'low', 20,
                                              "1d").min() and context.strat == TurtleshStrategy.CONST_LONG) or (
                        price_high >= data.history(context.object.asset, 'high', 20,
                                                   "1d").max() and context.strat == TurtleshStrategy.CONST_SHORT):
                    if TurtleshStrategy.DEBUG:
                        print("Exiting...")
                    context.object.sell_id = zipline.api.order_target(context.object.asset, 0)
                    context.strat = None

                self.sell_at_end(context, data)
            self.nn_record(context, data)

        return handle_data

    @staticmethod
    def tr(context, data):
        h = data.current(context.object.asset, 'high')
        l = data.current(context.object.asset, 'low')
        pdc = data.history(context.object.asset, 'close', 1, "1d").item()
        return max(h - l, h - pdc, pdc - l)

    @staticmethod
    def starting_n(context, data):
        h = data.history(context.object.asset, 'high', 20, "1d")
        l = data.history(context.object.asset, 'low', 20, "1d")
        pdc = data.history(context.object.asset, 'close', 21, "1d")
        trsum = 0.0
        for i in range(0, 20):
            trsum += max(h[i] - l[i], h[i] - pdc[i], pdc[i] - l[i])
        return trsum / 20.0

    @staticmethod
    def n(context, data):
        if context.first_run:
            context.first_run = False
            return TurtleshStrategy.starting_n(context, data)
        else:
            return (19 * context.pdn + TurtleshStrategy.tr(context, data)) / 20.0


class Test(unittest.TestCase):

    def test_turtlesh(self):
        t = TurtleshStrategy(datetime(2007, 1, 1, 0, 0, 0, 0, pytz.utc), 40, 10000, 'daily', None, 'X')
        self.assertIsInstance(t.run_backtest(), DataFrame)
