import csv
import numpy as np
import pandas as pd
from sklearn.model_selection import StratifiedShuffleSplit, StratifiedKFold
from sklearn.decomposition import PCA
from sklearn import manifold
from keras.utils import to_categorical
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import math
from collections import deque


class TrainingSet():

    def __init__(self, dir = '../../data/backtest.csv', eval_pct=0.2, num_folds=10):
        self.dir = dir
        self.eval_pct = eval_pct
        self.num_folds = num_folds
        self.k_folds = []
        self.input_size = None
        self.output_size = None

        self.eval_data = None
        self.type = None
        self.categories = None
        self.input_col = None
        self.output_col = None

    def getEvaluationData(self):
        if self.eval_data is None:
            self.parseRawFile()
        return self.eval_data

    def getKFolds(self):
        if len(self.k_folds) == 0:
            self.parseRawFile()
        return self.k_folds

    def getInputSize(self):
        if self.input_size is None:
            self.parseRawFile()
        return self.input_size

    def getOutputSize(self):
        if self.input_size is None:
            self.parseRawFile()
        return self.output_size

    def getType(self):
        if self.type is None:
            self.parseRawFile()
        return self.type

    def parseRawFile(self):

        num_lines = sum(1 for l in open(self.dir))
        skip_idx = [x for x in range(1, num_lines) if x > 1]
        df = pd.read_csv(self.dir, delimiter=',', delim_whitespace=False, header=0, skiprows=skip_idx)

        self.type = (df['TYPE'])[0]
        self.categories = (df['CATEGORIES'])[0]
        self.input_col = (df['INPUT_VECTOR_COL'])[0] + 1
        self.output_col = (df['OUTPUT_VECTOR_COL'])[0]
        data_width = (df['DATA_LENGTH'])[0]

        raw_data = np.empty((0, data_width))

        # (!!!) Normalizing here - correct if necessary
        # df = normalize_df(pd.read_csv(self.dir, delimiter=',', delim_whitespace=False, header=2),
        #                   ('EMA', 'LongEMA', 'RSI', 'PercentK', 'PercentD', 'SlowPercentD', 'Momentum', 'ROC',
        #                    'PercentR', 'AccumulationDistribution', 'ChalkinOscillator', 'CommoditiesChannelIndex',
        #                    'Disparity5', 'Disparity10'))
        df = pd.read_csv(self.dir, delimiter=',', delim_whitespace=False, header=2)
        for index, row in df.iterrows():
            raw_data = np.append(raw_data, [np.asarray(row)], axis=0)
        X = raw_data[:, self.input_col:self.output_col].astype('float64')
        y = raw_data[:, self.output_col:].astype('float64')

        if self.type == 'regression':
            self.input_size, self.output_size = X.shape[1], y.shape[1]
            y_w = self.output_size
            bins = np.linspace(-1, 1, 8)
            self.categories = 8
            y_binned = np.digitize(y, bins)
            y_binned = np.matmul(y_binned, np.power(np.full((y_w,1), 8), np.arange(y_w).reshape(y_w, 1)))
        else:
            y_exp = to_categorical(y, self.categories)
            y_binned = y
            y = y_exp
            self.input_size, self.output_size = X.shape[1], y.shape[1]

        #TODO: UNCOMMENT THIS!
        #self.findMDS(X, y_binned)

        split_train = StratifiedShuffleSplit(n_splits=1, train_size=1 - self.eval_pct, test_size=self.eval_pct)
        for train_index, test_index in split_train.split(X, y_binned):
            X_, y_, y_binned, = X[train_index], y[train_index], y_binned[train_index]
            X_val, y_val = X[test_index], y[test_index]
        self.eval_data = ((X_, y_),(X_val, y_val))

        split_kfold = StratifiedKFold(n_splits=self.num_folds, shuffle=True)
        for train_index, test_index in split_kfold.split(X_, y_binned):
            X_train, y_train, X_test, y_test = X_[train_index], y_[train_index], X_[test_index], y_[test_index]
            self.k_folds += [((X_train, y_train), (X_test, y_test))]

    def findMDS(self, X, y_binned):
        f, ((ax1, ax2), (ax3, ax4)) = plt.subplots(2, 2, sharex='col', sharey='row')
        ax1.set_title('Muti-Dimensional Scaling')
        #ax1.set_xlim([-1000, 1000])
        #ax1.set_ylim([-1000, 1000])

        ax2.set_title('ISO Map')
        #ax2.set_xlim([-1000, 1000])
        #ax2.set_ylim([-1000, 1000])

        ax3.set_title('t-SNE')
        #ax3.set_xlim([-1000, 1000])
        #ax3.set_ylim([-1000, 1000])

        ax4.set_title('PCA')
        #ax4.set_xlim([-1000, 1000])
        #ax4.set_ylim([-1000, 1000])


        mds = manifold.MDS(n_components=2, max_iter=100, n_init=1)
        isomap = manifold.Isomap(n_neighbors=5, n_components=2, max_iter=100)
        tsne = manifold.TSNE(n_components=2, init='pca', random_state=0)
        pca = PCA(n_components=2)

        palette = cm.rainbow(np.linspace(0, 1, self.categories))

        print("Starting transforms...")
        Y_man = mds.fit_transform(X)
        print("MDS transform complete.")
        Y_iso = isomap.fit_transform(X)
        print("ISOMap transform complete.")
        Y_tsne = tsne.fit_transform(X)
        print("TSNE transform complete.")
        Y_pca = pca.fit_transform(X)
        print("PCA transform complete.")

        for i in range(0, self.categories):
            loc = np.where(y_binned == i)[0]
            ax1.scatter(Y_man[loc, 0], Y_man[loc, 1], c=palette[i], marker='$'+str(i)+'$')
            ax2.scatter(Y_iso[loc, 0], Y_iso[loc, 1], c=palette[i], marker='$'+str(i)+'$')
            ax3.scatter(Y_tsne[loc, 0], Y_tsne[loc, 1], c=palette[i], marker='$'+str(i)+'$')
            ax4.scatter(Y_pca[loc, 0], Y_pca[loc, 1], c=palette[i], marker='$'+str(i)+'$')
        plt.show()

    def sampleDataSTDev(self, array):
        input_sample = array[self.input_col:self.output_col - 1]
        output_sample = array[self.output_col:]

        input_mean = np.average(input_sample, axis=0)
        input_data = np.square(input_sample) / np.size(input_sample, 0)

        output_mean = np.average(output_sample, axis=0)
        output_data = np.square(output_sample) / np.size(output_sample, 0)

        return np.sum(input_data, axis=0) - input_mean, np.sum(output_data, axis=0) - output_mean



