from math import ceil

import matplotlib as plt
from sklearn.metrics import confusion_matrix, f1_score, roc_auc_score, roc_curve
from mlxtend.plotting import plot_confusion_matrix
from keras.utils import to_categorical
import numpy as np
import matplotlib.pyplot as plt
import sys
from nn.neural_networks.manifest import Manifest

class NeuralNetwork():
    def __init__(self, type):
        self.type = type
        self.model = None
        self.kfold_results = []
        self.eval_result = None
        self.ts = None

    def train_model(self, ts):
        self.ts = ts
        eval_data = ts.getEvaluationData()
        kfold_data = ts.getKFolds()
        type = ts.getType()

        if self.type != type:
            print("Error: Dataset type {} does not match neural network type {}.".format(type, self.type))
            sys.exit()

        # self.build_network(ts.getInputSize(), ts.getOutputSize())
        #
        # ((X_train, y_train), (X_test, y_test)) = kfold_data[1]
        # self.train_once(X_train, y_train, X_test, y_test)

        X_eval, y_eval = (eval_data[1])[0], (eval_data[1])[1]
        for ((X_train, y_train), (X_test, y_test)) in kfold_data:
            self.build_network(ts.getInputSize(), ts.getOutputSize())
            self.kfold_results += [self.train_once(X_train, y_train, X_test, y_test)]
            eval_result = self.model.predict(X_eval)
            self.validate(X_eval, y_eval, eval_result)
            self.build_confusion_tables(eval_result, y_eval)

    def train_once(self, X_train, y_train, X_val, y_val):
        history = self.model.fit(X_train, y_train,
                            validation_data=(X_val, y_val),
                            #do i overfit when i set it to 2000?
                            epochs=1000,
                            #epochs=100,
                            batch_size=50)

        return history

    def build_network(self, input_shape, output_shape):
        pass

    def build_confusion_tables(self, y_predict, y_actual):
        pass

    def validate(self, X_eval, y_eval, eval_result):
        pass

    def save(self, weights_path, manifest_path):
        self.model.save_weights(weights_path)
        manifest = Manifest(self.ts.dir, weights_path, type(self))
        manifest.save(manifest_path)

class RegressionNeuralNetwork(NeuralNetwork):

    def __init__(self):
        super(RegressionNeuralNetwork, self).__init__('regression')

    def build_network(self, input_shape, output_shape):
        pass

    def build_confusion_tables(self, y_predict, y_actual):
        bins = np.linspace(-1, 1, 21)
        p_col = np.digitize(y_predict, bins)
        a_col = np.digitize(y_actual, bins)
        confusionMatrix = confusion_matrix(a_col, p_col)
        binary = np.array(confusionMatrix)
        fig, ax = plot_confusion_matrix(conf_mat=binary)
        plt.show()

    def validate(self, X_eval, y_eval, eval_result):
        score = self.model.evaluate(X_eval, y_eval, batch_size=50)
        print("Validation set accuracy: ", score[1])


class ClassificationNeuralNetwork(NeuralNetwork):

    def __init__(self):
        super(ClassificationNeuralNetwork, self).__init__('classification')

    def build_confusion_tables(self, y_predict, y_actual):
        confusionMatrix = confusion_matrix(y_actual.argmax(axis=1), y_predict.argmax(axis=1))
        binary = np.array(confusionMatrix)
        fig, ax = plot_confusion_matrix(conf_mat=binary)
        plt.show()

    def validate(self, X_eval, y_eval, eval_result):

        def get_subplots():
            if self.ts.categories < 4:
                columns = self.ts.categories
                rows = 1
            else:
                columns = 4
                rows = ceil(self.ts.categories / 4)
            f, ax = plt.subplots(rows, columns, sharex='col', sharey='row')
            return ax.flatten()

        subplots = get_subplots()
        score = self.model.evaluate(X_eval, y_eval, batch_size=50)
        y_true, y_predict = y_eval.argmax(axis=1), eval_result.argmax(axis=1)

        f1_macro = f1_score(y_true, y_predict, average='macro')
        f1_weighted = f1_score(y_true, y_predict, average='weighted')

        auroc_macro = []
        auroc_weighted = []
        ytrue_cat, ypredict_cat = to_categorical(y_true, self.ts.categories), to_categorical(y_predict, self.ts.categories)
        for i in range(0, self.ts.categories):
            auroc_macro += [roc_auc_score(ytrue_cat[:, i], ypredict_cat[:, i], average='macro')]
            auroc_weighted += [roc_auc_score(ytrue_cat[:, i], ypredict_cat[:, i], average='weighted')]
            x, y, _ = roc_curve(ytrue_cat[:, i], ypredict_cat[:, i])
            #"Class " + i + " ROC curve"
            subplots[i].set_title("Class %d ROC curve" % i)
            subplots[i].set_xlim([0.0, 1.0])
            subplots[i].set_ylim([0.0, 1.05])
            subplots[i].set_xlabel('False Positive Rate')
            subplots[i].set_ylabel('True Positive Rate')
            subplots[i].plot(x, y, color='darkorange', lw=2, label = 'ROC curve (area = %0.2f)' % auroc_macro[-1])
            subplots[i].legend(loc="lower right")
            subplots[i].plot([0, 1], [0, 1], color='navy', lw=2, linestyle='--')

        diff = np.absolute(y_true - y_predict)
        bad = np.where(diff >= 2)[0].size

        print("Validation set accuracy: ", score[1])
        print("F1-Score Unweighted: ", f1_macro)
        print("F1-Score Weighted: ", f1_weighted)
        print("Area Under ROC Curve Unweighted: ", auroc_macro)
        print("Area Under ROC Curve Weighted: ", auroc_weighted)
        print(round(bad/diff.size *100.0, 2), "% of elements were grossly incorrect")

        plt.show()


