import csv
import numpy as np
import pandas as pd
from sklearn.model_selection import StratifiedShuffleSplit, StratifiedKFold
from sklearn.decomposition import PCA
from sklearn import manifold
from keras.utils import to_categorical
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import math
from keras import optimizers
from collections import deque
from keras.models import Sequential
from keras.layers import Dense
from keras.layers import LSTM
from sklearn.preprocessing import MinMaxScaler
from sklearn.metrics import mean_squared_error

class TrainingSet():

    def __init__(self, dir = '../../data/backtest.csv', eval_pct=0.2, num_folds=10):
        self.dir = dir
        self.eval_pct = eval_pct
        self.num_folds = num_folds
        self.k_folds = []
        self.input_size = None
        self.output_size = None

        self.eval_data = None
        self.type = None
        self.categories = None
        self.input_col = None
        self.output_col = None

    def getEvaluationData(self):
        if self.eval_data is None:
            self.parseRawFile()
        return self.eval_data

    def getKFolds(self):
        if len(self.k_folds) == 0:
            self.parseRawFile()
        return self.k_folds

    def getInputSize(self):
        if self.input_size is None:
            self.parseRawFile()
        return self.input_size

    def getOutputSize(self):
        if self.output_size is None:
            self.parseRawFile()
        return self.output_size

    def getType(self):
        if self.type is None:
            self.parseRawFile()
        return self.type

    def parseRawFile(self):

        dir = self.dir
        num_lines = sum(1 for l in open(dir))
        skip_idx = [x for x in range(1, num_lines) if x > 1]
        df = pd.read_csv(dir, delimiter=',', delim_whitespace=False, header=0, skiprows=skip_idx)

        self.input_col = (df['INPUT_VECTOR_COL'])[0]
        self.output_col = (df['OUTPUT_VECTOR_COL'])[0] - 1
        data_width = self.output_col-self.input_col

        df = pd.read_csv(dir, delimiter=',', delim_whitespace=False, header=2)

        sample_count = int(df.iloc[-1]['SID']) + 1
        x_list = [np.empty((0, self.output_col-self.input_col)) for i in range(0, sample_count)]
        y_list = [None for i in range(0, sample_count)]

        for index, row in df.iterrows():
            sid = int(row[0])
            print(sid)
            x_list[sid] = np.append(x_list[sid], [np.asarray(row[self.input_col:self.output_col])], axis=0)
            if y_list[sid] == None: y_list[sid] = row[self.output_col+1]

        X = np.zeros((x_list[0].shape[0], x_list[0].shape[1], 0))
        for elt in x_list:
            X = np.dstack((X, elt))
        y = np.asarray(y_list)

        #NOTE: I will not support regression for the LSTM model
        X = np.reshape(X, (X.shape[2], X.shape[0], X.shape[1]))
        y_binned = y
        y = to_categorical(y)

        split_train = StratifiedShuffleSplit(n_splits=1, train_size=1 - self.eval_pct, test_size=self.eval_pct)
        for train_index, test_index in split_train.split(X, y_binned):
            X_, y_, y_binned, = X[train_index,:,:], y[train_index], y_binned[train_index]
            X_val, y_val = X[test_index,:,:], y[test_index]
        self.eval_data = ((X_, y_), (X_val, y_val))

        data_split = int(0.75 * sample_count)
        x_train, x_test = X[:,:,:data_split], X[:,:,data_split:]
        y_train, y_test = y_data[:data_split], y_data[data_split:]

        x_train = np.reshape(x_train, (x_train.shape[2], x_train.shape[0], x_train.shape[1]))
        x_test = np.reshape(x_test, (x_test.shape[2], x_test.shape[0], x_test.shape[1]))
        print(x_train.shape)
        print(x_test.shape)

        scaler = MinMaxScaler(feature_range=(0, 1))

        model = Sequential()
        model.add(LSTM(20, return_sequences=True, input_shape=(x_train.shape[1], x_train.shape[2])))
        model.add(LSTM(20, return_sequences=True, input_shape=(x_train.shape[1], x_train.shape[2])))
        model.add(LSTM(20, input_shape=(x_train.shape[1], x_train.shape[2])))
        model.add(Dense(2, kernel_initializer='glorot_uniform', activation='softmax', use_bias=True))
        model.compile(loss='categorical_crossentropy', optimizer='sgd')
        print(y_train)
        model.fit(x_train, y_train, epochs=100, batch_size=1, verbose=2)

        trainPredict = model.predict(x_train)
        testPredict = model.predict(x_test)

        print(testPredict, y_test)
        trainPredict = scaler.inverse_transform(trainPredict)
        y_train = scaler.inverse_transform([y_train])
        testPredict = scaler.inverse_transform(testPredict)
        y_test = scaler.inverse_transform([y_test])
        # calculate root mean squared error
        trainScore = math.sqrt(mean_squared_error(y_train[0], trainPredict[:, 0]))
        print('Train Score: %.2f RMSE' % (trainScore))
        testScore = math.sqrt(mean_squared_error(y_test[0], testPredict[:, 0]))
        print('Test Score: %.2f RMSE' % (testScore))

if __name__ == '__main__':
    TrainingSet ts = TrainingSet()
    ts.parseRawFile("../../data/mac_tonight_2018-07-22_14-09-13.csv")

