# Copyright [2018, +inf), Zoraver Kang (zkang@wpi.edu) and Remy Kaldawy
# At this time, the system does not support trading in real time
# Additionally, only daily bars are currently supported, and the historical period is the same as the trading period

from collections import deque
from datetime import datetime, timedelta

import numpy as np
import pytz

from bt.backtest.cnn.data_generate import evaluate_performance, GeneratorStrategy
from nn.cnn.neural_networks.networks import *  # needed for Manifest.load to work
from nn.neural_networks.manifest import Manifest
from bt.strategies.strategies import SimpleEMAStrategy
from bt.strategies.strategy_params import DailyStrategyParameters


# This class was created to store the last n prices of a specific asset. This was done to facilitate trading periods not
# equal to the historical period, which is defined during the training of the model.
# This class may no longer be needed - the program will only have to change slightly for it to be removed entirely.
class FixedSizeSelfPruningQueue(object):

    def __init__(self, size):
        self.max_size = size
        self.storage = deque()

    def __iter__(self):
        return self.storage.__iter__()

    def enqueue(self, item):
        if len(self.storage) >= self.max_size:
            self.storage.popleft()
        self.storage.append(item)


# Extends GeneratorStrategy to use a FixedSizeSelfPruningQueue to store pricing data.
class AdaptableGeneratorStrategy(GeneratorStrategy):

    def __init__(self, params, asset, indicators, queue):
        super(AdaptableGeneratorStrategy, self).__init__(params, asset, indicators)
        self.queue = queue

    def handle_data(self):
        def handle_data(context, data):
            self.queue.enqueue(data.current(context.asset, 'price'))

        return handle_data


class Trader(object):

    def __init__(self, start, end, historical_period, trading_period, principal, asset, model, strategies,
                 indicators, debug=False):
        self.start = start
        self.end = end
        self.historical_period = historical_period
        self.trading_period = trading_period
        self.principal = principal
        self.asset = asset
        self.model = model  # load_model(model_dir, custom_objects={'FixedLeakyReLU': FixedLeakyReLU})
        self.strategies = strategies
        self.indicators = indicators
        self.debug = debug

        self.current = self.start
        self.pricing = FixedSizeSelfPruningQueue(historical_period)
        self.generator = AdaptableGeneratorStrategy(
            DailyStrategyParameters(self.start - timedelta(days=self.historical_period),
                                    self.historical_period, self.principal, 'daily',
                                    None, self.asset),
            self.asset, self.indicators, self.pricing)

    def run(self):
        while self.current < self.end:
            weights = self.model.predict_classes(self.get_input_vector())
            if self.debug:
                print('current, gcurrent', self.current, self.current.date().weekday(),
                      self.generator.params.start, self.generator.params.start.date().weekday())
                print('weights', weights)
            self.allocate_funds(weights)
            for strategy in self.strategies:
                strategy.run_backtest()
            self.collect_funds()
            for strategy in self.strategies:
                strategy.increment_timedelta()
            self.current += timedelta(days=self.trading_period)
            self.generator.params.update(start=self.current - timedelta(days=self.historical_period))
        print('trading ended, principal', self.principal)

    def get_input_vector(self):
        vector = []
        self.generator.run_backtest()
        for price in self.pricing:
            vector.append(price)
        for indicator_str in self.indicators:
            vector.append(self.generator.indicator_values[indicator_str])
        for strategy in self.strategies:
            strategy.params.update(principal=self.principal)  # ensure consistent principal for evaluation
            strategy.run_backtest()
            vector.extend(evaluate_performance([strategy.result]))
        if self.debug:
            print('vector', len(vector), vector)
        return np.array([vector])  # really dumb that this works...

    def allocate_funds(self, weights):
        # Using a basic "averaging" allocation strategy
        # Also assuming that the weights list and the strategies list are ordered in the same way - need to verify!
        weight_total = 7 * len(weights)
        allocated_total = 0

        for i in range(0, len(self.strategies)):
            principal = 1 + weights[i] * (self.principal - len(self.strategies)) / weight_total
            allocated_total += principal
            self.strategies[i].params.update(principal=principal)

        self.principal -= allocated_total
        if self.debug:
            print('principal after allocation, funds allocated', self.principal, allocated_total)

    def collect_funds(self):
        for strategy in self.strategies:
            res = strategy.result.last('1D')  # will need to change once minute bars are supported
            # ending_value used to cover the edge case where the stocks are not actually sold at the end...
            self.principal += res['ending_cash'].values[0] + res['ending_value'].values[0]
        if self.debug:
            print('principal after funds collected', self.principal)


if __name__ == '__main__':
    # ts = TrainingSet('../data/doodeedoodee.csv')
    # cnn = CategoricalNeuralNetwork()
    # cnn.build_network(ts.getInputSize(), ts.getOutputSize())
    # cnn.model.load_weights('../data/models/test_weights.h5')

    # indicators = ['EMA', 'LongEMA', 'RSI', 'PercentK', 'PercentD', 'SlowPercentD', 'Momentum', 'ROC', 'PercentR',
    #              'AccumulationDistribution', 'ChalkinOscillator', 'CommoditiesChannelIndex', 'Disparity5',
    #              'Disparity10']
    manifest = Manifest.load('data/models/test_manifest.pickle')
    start = datetime(2007, 1, 1, 0, 0, 0, 0, pytz.utc)
    trader = Trader(start, datetime(2017, 1, 1, 0, 0, 0, 0, pytz.utc), 40, 40,
                    1000.0, 'AAPL', manifest.get_network().model,
                    [SimpleEMAStrategy(start - timedelta(days=40), 40, 1000.0,
                                       'daily', None, 'AAPL')], manifest.indicators, True)
    trader.run()
